package com.github.zengfr.easymodbus4j.channel;

import com.github.zengfr.easymodbus4j.ModbusConfs;
import com.github.zengfr.easymodbus4j.ModbusConsts;
import com.github.zengfr.easymodbus4j.codec.ping.ModbusPingDecoder;
import com.github.zengfr.easymodbus4j.codec.ping.ModbusPingEncoder;
import com.github.zengfr.easymodbus4j.handler.ModbusHeartbeatHandler;
import com.github.zengfr.easymodbus4j.handler.ModbusLoggingHandler;
import com.github.zengfr.easymodbus4j.protocol.tcp.ModbusFrame;
import io.netty.channel.ChannelHandler;
import io.netty.channel.ChannelHandler.Sharable;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.SimpleChannelInboundHandler;
import io.netty.channel.socket.SocketChannel;
import io.netty.handler.logging.LogLevel;
import io.netty.handler.timeout.IdleStateHandler;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;

@Sharable
public abstract class ModbusChannelInitializer extends ChannelInitializer<SocketChannel> {
    private static final InternalLogger logger = InternalLoggerFactory.getInstance(ModbusChannelInitializer.class);

    private final SimpleChannelInboundHandler<ModbusFrame> handler;

    protected final Boolean isSlave;

    public ModbusChannelInitializer(Boolean isSlave, SimpleChannelInboundHandler<ModbusFrame> handler) {
        this.isSlave = isSlave;
        this.handler = handler;
    }

    protected void initChannel(SocketChannel ch) throws Exception {
        logger.info(String.format("initChannel->isSlave:%s", this.isSlave));
        if (ch.localAddress() != null || ch.remoteAddress() != null)
            logger.info(String.format("initChannel->%s,%s", ch.localAddress(), ch.remoteAddress()));
        ChannelPipeline pipeline = ch.pipeline();
        initPipeline4Logging(pipeline);
        initPipeline4Ping(pipeline);
        initPipelinePreProcessHandler(pipeline);
        initPipeline4ProcessHandler(pipeline);
        initPipelinePostProcessHandler(pipeline);
        initPipeline4HeartbeatIdle(pipeline);
    }

    protected void initPipeline4Logging(ChannelPipeline pipeline) {
        if (isShowDebugLog(this.isSlave.booleanValue()))
            pipeline.addLast("logging", new ModbusLoggingHandler(LogLevel.DEBUG));
    }

    protected void initPipeline4Ping(ChannelPipeline pipeline) {
        Charset charset = StandardCharsets.UTF_8;
        pipeline.addLast("pingDecoder", new ModbusPingDecoder(ModbusConsts.HEARTBEAT, charset));
        pipeline.addLast("pingEncoder", new ModbusPingEncoder(charset));
    }

    protected void initPipeline4ProcessHandler(ChannelPipeline pipeline) {
        if (this.handler != null)
            if (this.isSlave.booleanValue()) {
                pipeline.addLast("requestHandler", this.handler);
            } else {
                pipeline.addLast("responseHandler", this.handler);
            }
    }

    protected void initPipeline4HeartbeatIdle(ChannelPipeline pipeline) {
        int timeOut = getIdleTimeout();
        if (timeOut > 1) {
            pipeline.addLast("idleStateHandler", new IdleStateHandler(timeOut, 0, 0));
            pipeline.addLast("heartbeatHandler", new ModbusHeartbeatHandler(this.isSlave.booleanValue()));
        }
    }

    protected static int getIdleTimeout() {
        return ModbusConfs.IDLE_TIMEOUT_SECOND;
    }

    protected static boolean isShowDebugLog(boolean isSlave) {
        return ((isSlave && ModbusConfs.SLAVE_SHOW_DEBUG_LOG) || (ModbusConfs.MASTER_SHOW_DEBUG_LOG && !isSlave));
    }

    protected abstract void initPipelinePreProcessHandler(ChannelPipeline paramChannelPipeline);

    protected abstract void initPipelinePostProcessHandler(ChannelPipeline paramChannelPipeline);
}
