package com.github.zengfr.easymodbus4j.common.util;

import com.github.zengfr.easymodbus4j.common.RegisterOrder;

import java.nio.ByteBuffer;
import java.util.Random;

public class RegistersUtil {
    public static int[] getRandomRegisters(int length, int minInt, int maxInt, Random random) {
        int[] registers = new int[length];
        for (int i = 0; i < registers.length; i++)
            registers[i] = minInt + random.nextInt(maxInt);
        return registers;
    }

    public static byte[] toByteArray(int value) {
        byte[] result = new byte[2];
        result[1] = (byte) (value >> 8);
        result[0] = (byte) value;
        return result;
    }

    public static byte[] toByteArrayShort(short value) {
        return ByteBuffer.allocate(2).putShort(value).array();
    }

    public static byte[] toByteArrayInt(int value) {
        return ByteBuffer.allocate(4).putInt(value).array();
    }

    public static byte[] toByteArrayLong(long value) {
        return ByteBuffer.allocate(8).putLong(value).array();
    }

    public static byte[] toByteArrayDouble(double value) {
        return ByteBuffer.allocate(8).putDouble(value).array();
    }

    public static byte[] toByteArrayFloat(float value) {
        return ByteBuffer.allocate(4).putFloat(value).array();
    }

    public static float convertRegistersToFloat(int[] registers) throws IllegalArgumentException {
        if (registers.length != 2)
            throw new IllegalArgumentException("Input Array length invalid");
        int highRegister = registers[1];
        int lowRegister = registers[0];
        byte[] highRegisterBytes = toByteArray(highRegister);
        byte[] lowRegisterBytes = toByteArray(lowRegister);
        byte[] floatBytes = {highRegisterBytes[1], highRegisterBytes[0], lowRegisterBytes[1], lowRegisterBytes[0]};
        return ByteBuffer.wrap(floatBytes).getFloat();
    }

    public static double convertRegistersToDouble(int[] registers) throws IllegalArgumentException {
        if (registers.length != 4)
            throw new IllegalArgumentException("Input Array length invalid");
        byte[] highRegisterBytes = toByteArray(registers[3]);
        byte[] highLowRegisterBytes = toByteArray(registers[2]);
        byte[] lowHighRegisterBytes = toByteArray(registers[1]);
        byte[] lowRegisterBytes = toByteArray(registers[0]);
        byte[] doubleBytes = {highRegisterBytes[1], highRegisterBytes[0], highLowRegisterBytes[1], highLowRegisterBytes[0], lowHighRegisterBytes[1], lowHighRegisterBytes[0], lowRegisterBytes[1], lowRegisterBytes[0]};
        return ByteBuffer.wrap(doubleBytes).getDouble();
    }

    public static double convertRegistersToDouble(int[] registers, RegisterOrder registerOrder) throws IllegalArgumentException {
        if (registers.length != 4)
            throw new IllegalArgumentException("Input Array length invalid");
        int[] swappedRegisters = {registers[0], registers[1], registers[2], registers[3]};
        if (registerOrder == RegisterOrder.HighLow)
            swappedRegisters = new int[]{registers[3], registers[2], registers[1], registers[0]};
        return convertRegistersToDouble(swappedRegisters);
    }

    public static float convertRegistersToFloat(int[] registers, RegisterOrder registerOrder) throws IllegalArgumentException {
        int[] swappedRegisters = {registers[0], registers[1]};
        if (registerOrder == RegisterOrder.HighLow)
            swappedRegisters = new int[]{registers[1], registers[0]};
        return convertRegistersToFloat(swappedRegisters);
    }

    public static long convertRegistersToLong(int[] registers) throws IllegalArgumentException {
        if (registers.length != 4)
            throw new IllegalArgumentException("Input Array length invalid");
        byte[] highRegisterBytes = toByteArray(registers[3]);
        byte[] highLowRegisterBytes = toByteArray(registers[2]);
        byte[] lowHighRegisterBytes = toByteArray(registers[1]);
        byte[] lowRegisterBytes = toByteArray(registers[0]);
        byte[] longBytes = {highRegisterBytes[1], highRegisterBytes[0], highLowRegisterBytes[1], highLowRegisterBytes[0], lowHighRegisterBytes[1], lowHighRegisterBytes[0], lowRegisterBytes[1], lowRegisterBytes[0]};
        return ByteBuffer.wrap(longBytes).getLong();
    }

    public static long convertRegistersToLong(int[] registers, RegisterOrder registerOrder) throws IllegalArgumentException {
        if (registers.length != 4)
            throw new IllegalArgumentException("Input Array length invalid");
        int[] swappedRegisters = {registers[0], registers[1], registers[2], registers[3]};
        if (registerOrder == RegisterOrder.HighLow)
            swappedRegisters = new int[]{registers[3], registers[2], registers[1], registers[0]};
        return convertRegistersToLong(swappedRegisters);
    }

    public static int convertRegistersToInt(int[] registers) throws IllegalArgumentException {
        if (registers.length != 2)
            throw new IllegalArgumentException("Input Array length invalid");
        int highRegister = registers[1];
        int lowRegister = registers[0];
        byte[] highRegisterBytes = toByteArray(highRegister);
        byte[] lowRegisterBytes = toByteArray(lowRegister);
        byte[] doubleBytes = {highRegisterBytes[1], highRegisterBytes[0], lowRegisterBytes[1], lowRegisterBytes[0]};
        return ByteBuffer.wrap(doubleBytes).getInt();
    }

    public static int convertRegistersToInt(int[] registers, RegisterOrder registerOrder) throws IllegalArgumentException {
        int[] swappedRegisters = {registers[0], registers[1]};
        if (registerOrder == RegisterOrder.HighLow)
            swappedRegisters = new int[]{registers[1], registers[0]};
        return convertRegistersToInt(swappedRegisters);
    }

    public static int[] convertFloatToRegisters(float floatValue) {
        byte[] floatBytes = toByteArrayFloat(floatValue);
        byte[] highRegisterBytes = {0, 0, floatBytes[0], floatBytes[1]};
        byte[] lowRegisterBytes = {0, 0, floatBytes[2], floatBytes[3]};
        int[] returnValue = {ByteBuffer.wrap(lowRegisterBytes).getInt(), ByteBuffer.wrap(highRegisterBytes).getInt()};
        return returnValue;
    }

    public static int[] convertFloatToRegisters(float floatValue, RegisterOrder registerOrder) {
        int[] registerValues = convertFloatToRegisters(floatValue);
        int[] returnValue = registerValues;
        if (registerOrder == RegisterOrder.HighLow)
            returnValue = new int[]{registerValues[1], registerValues[0]};
        return returnValue;
    }

    public static int[] convertIntToRegisters(int intValue) {
        byte[] doubleBytes = toByteArrayInt(intValue);
        byte[] highRegisterBytes = {0, 0, doubleBytes[0], doubleBytes[1]};
        byte[] lowRegisterBytes = {0, 0, doubleBytes[2], doubleBytes[3]};
        int[] returnValue = {ByteBuffer.wrap(lowRegisterBytes).getInt(), ByteBuffer.wrap(highRegisterBytes).getInt()};
        return returnValue;
    }

    public static int[] convertIntToRegisters(int intValue, RegisterOrder registerOrder) {
        int[] registerValues = convertIntToRegisters(intValue);
        int[] returnValue = registerValues;
        if (registerOrder == RegisterOrder.HighLow)
            returnValue = new int[]{registerValues[1], registerValues[0]};
        return returnValue;
    }

    public static int[] convertLongToRegisters(long longValue) {
        byte[] doubleBytes = toByteArrayLong(longValue);
        byte[] highhighRegisterBytes = {0, 0, doubleBytes[0], doubleBytes[1]};
        byte[] highlowRegisterBytes = {0, 0, doubleBytes[2], doubleBytes[3]};
        byte[] lowHighRegisterBytes = {0, 0, doubleBytes[4], doubleBytes[5]};
        byte[] lowlowRegisterBytes = {0, 0, doubleBytes[6], doubleBytes[7]};
        int[] returnValue = {ByteBuffer.wrap(lowlowRegisterBytes).getInt(), ByteBuffer.wrap(lowHighRegisterBytes).getInt(), ByteBuffer.wrap(highlowRegisterBytes).getInt(), ByteBuffer.wrap(highhighRegisterBytes).getInt()};
        return returnValue;
    }

    public static int[] convertLongToRegisters(long longValue, RegisterOrder registerOrder) {
        int[] registerValues = convertLongToRegisters(longValue);
        int[] returnValue = registerValues;
        if (registerOrder == RegisterOrder.HighLow)
            returnValue = new int[]{registerValues[3], registerValues[2], registerValues[1], registerValues[0]};
        return returnValue;
    }

    public static int[] convertDoubleToRegisters(double doubleValue) {
        byte[] doubleBytes = toByteArrayDouble(doubleValue);
        byte[] highhighRegisterBytes = {0, 0, doubleBytes[0], doubleBytes[1]};
        byte[] highlowRegisterBytes = {0, 0, doubleBytes[2], doubleBytes[3]};
        byte[] lowHighRegisterBytes = {0, 0, doubleBytes[4], doubleBytes[5]};
        byte[] lowlowRegisterBytes = {0, 0, doubleBytes[6], doubleBytes[7]};
        int[] returnValue = {ByteBuffer.wrap(lowlowRegisterBytes).getInt(), ByteBuffer.wrap(lowHighRegisterBytes).getInt(), ByteBuffer.wrap(highlowRegisterBytes).getInt(), ByteBuffer.wrap(highhighRegisterBytes).getInt()};
        return returnValue;
    }

    public static int[] convertDoubleToRegisters(double doubleValue, RegisterOrder registerOrder) {
        int[] registerValues = convertDoubleToRegisters(doubleValue);
        int[] returnValue = registerValues;
        if (registerOrder == RegisterOrder.HighLow)
            returnValue = new int[]{registerValues[3], registerValues[2], registerValues[1], registerValues[0]};
        return returnValue;
    }

    public static String convertRegistersToString(int[] registers, int offset, int stringLength) {
        byte[] result = new byte[stringLength];
        byte[] registerResult = new byte[2];
        for (int i = 0; i < stringLength / 2; i++) {
            registerResult = toByteArray(registers[offset + i]);
            result[i * 2] = registerResult[0];
            result[i * 2 + 1] = registerResult[1];
        }
        return new String(result);
    }

    public static int[] convertStringToRegisters(String stringToconvert) {
        byte[] array = stringToconvert.getBytes();
        int[] returnarray = new int[stringToconvert.length() / 2 + stringToconvert.length() % 2];
        for (int i = 0; i < returnarray.length; i++) {
            returnarray[i] = array[i * 2];
            if (i * 2 + 1 < array.length)
                returnarray[i] = returnarray[i] | array[i * 2 + 1] << 8;
        }
        return returnarray;
    }
}
