package com.github.zengfr.easymodbus4j.handler;

import com.github.zengfr.easymodbus4j.common.util.ScheduledUtil;
import com.github.zengfr.easymodbus4j.event.SyncStateEvent;
import com.github.zengfr.easymodbus4j.logging.ChannelLogger;
import com.google.common.collect.Maps;
import io.netty.channel.Channel;
import io.netty.channel.ChannelDuplexHandler;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelPromise;
import io.netty.util.Attribute;
import io.netty.util.AttributeKey;

import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class ModbusChannelSyncHandler extends ChannelDuplexHandler {
    private static final ChannelLogger logger = ChannelLogger.getLogger(ModbusChannelSyncHandler.class);

    private static final TimeUnit unit = TimeUnit.MILLISECONDS;

    private static final Random random = new Random();

    private final Map<Short, Semaphore> semaphoreMap;

    private final Map<Short, AtomicInteger> waitingMap;

    private final int permits;

    private final boolean isAutoAcquire;

    private final boolean isAutoRelease;

    private final long timeoutMillis;

    public ModbusChannelSyncHandler(long timeoutMillis) {
        this(1, false, false, timeoutMillis);
    }

    public ModbusChannelSyncHandler(boolean isAutoAcquire, boolean isAutoRelease, long timeoutMillis) {
        this(1, isAutoAcquire, isAutoRelease, timeoutMillis);
    }

    public ModbusChannelSyncHandler(int permits, boolean isAutoAcquire, boolean isAutoRelease, long timeoutMillis) {
        this.permits = permits;
        this.isAutoAcquire = isAutoAcquire;
        this.isAutoRelease = isAutoRelease;
        this.timeoutMillis = timeoutMillis;
        this.semaphoreMap = Maps.newConcurrentMap();
        this.waitingMap = Maps.newConcurrentMap();
        for (int i = 0; i <= 43; i++) {
            this.waitingMap.put((short) i, new AtomicInteger(0));
            this.semaphoreMap.put((short) i, new Semaphore(this.permits, true));
        }
        Runnable runnable = () -> {
            Set<Short> keys = this.waitingMap.keySet();
            for (Short key : keys) {
                AtomicInteger waiting = this.waitingMap.get(key);
                if (waiting.get() > 0)
                    waiting.decrementAndGet();
            }
        };
        ScheduledUtil.scheduleWithFixedDelay(runnable, this.timeoutMillis);
    }

    public static void fireUserSyncStateEvent(ChannelHandlerContext ctx, Short funcCode, boolean read, boolean write) throws Exception {
        SyncStateEvent evt = new SyncStateEvent(funcCode, read, write);
        ctx.fireUserEventTriggered(evt);
    }

    public void write(ChannelHandlerContext ctx, Object msg, ChannelPromise promise) throws Exception {
        if (this.isAutoAcquire) ;
        super.write(ctx, msg, promise);
    }

    public void read(ChannelHandlerContext ctx) throws Exception {
        super.read(ctx);
    }

    public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception {
        logger.debug(ctx.channel(), "channelRead");
        super.channelRead(ctx, msg);
        if (this.isAutoRelease) ;
    }

    public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
        if (evt instanceof SyncStateEvent) {
            SyncStateEvent event = (SyncStateEvent) evt;
            processSyncStateEvent(ctx.channel(), event);
        } else {
            super.userEventTriggered(ctx, evt);
        }
    }

    protected void processSyncStateEvent(Channel channel, SyncStateEvent event) {
        if (event.read)
            tryRelease(channel, event.funcCode);
        if (event.write)
            tryAcquire(channel, event.funcCode);
    }

    public boolean tryAcquire(Channel channel, Short funcCode) {
        AtomicInteger waiting = this.waitingMap.get(funcCode);
        int waitingCount = waiting.getAndIncrement();
        if (waitingCount < 0) {
            waitingCount = 0;
            waiting.set(waitingCount);
        }
        return tryAcquire(channel, funcCode, waitingCount);
    }

    protected boolean tryAcquire(Channel channel, Short funcCode, int waitingCount) {
        boolean success = false;
        int key = buildRandom();
        try {
            showLog(channel, funcCode, key, "tryAcquire:" + waitingCount);
            success = this.semaphoreMap.get(funcCode).tryAcquire(this.timeoutMillis * (waitingCount + 1), unit);
            showLog(channel, funcCode, key, "acquire:" + waitingCount + " " + success);
        } catch (InterruptedException ex) {
            logger.error(channel, ex);
        }
        return success;
    }

    public void tryRelease(Channel channel, Short funcCode) {
        int key = buildRandom();
        AttributeKey<Boolean> attrKey = AttributeKey.valueOf("isFrameIgnored");
        Attribute<Boolean> attr = channel.attr(attrKey);
        Boolean isFrameIgnored = attr.get();
        showLog(channel, funcCode, key, "isFrameIgnored:" + isFrameIgnored);
        if (isFrameIgnored == null || isFrameIgnored.equals(Boolean.valueOf(false))) {
            showLog(channel, funcCode, key, "tryRelease");
            Semaphore semaphore = this.semaphoreMap.get(funcCode);
            if (semaphore.availablePermits() < this.permits) {
                semaphore.release();
                AtomicInteger waiting = this.waitingMap.get(funcCode);
                waiting.decrementAndGet();
                showLog(channel, funcCode, key, "release");
            }
        }
    }

    private static int buildRandom() {
        return random.nextInt();
    }

    private void showLog(Channel channel, Short funcCode, int key, String mothod) {
        Semaphore semaphore = this.semaphoreMap.get(funcCode);
        logger.debug(channel, String.format("%s,%s,%s,avail:%s,q:%s", Integer.valueOf(key), funcCode, mothod, Integer.valueOf(semaphore.availablePermits()), Integer.valueOf(semaphore.getQueueLength())));
    }
}
