package com.github.zengfr.easymodbus4j.handler;

import com.github.zengfr.easymodbus4j.ModbusConsts;
import com.github.zengfr.easymodbus4j.logging.ChannelLogger;
import com.google.common.collect.Maps;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandler.Sharable;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import io.netty.handler.timeout.IdleState;
import io.netty.handler.timeout.IdleStateEvent;
import io.netty.util.CharsetUtil;
import io.netty.util.concurrent.GenericFutureListener;

import java.util.Map;

@Sharable
public class ModbusHeartbeatHandler extends ChannelInboundHandlerAdapter {
    private static final ChannelLogger logger = ChannelLogger.getLogger(ModbusHeartbeatHandler.class);

    private static final ByteBuf HEARTBEAT = Unpooled.unreleasableBuffer(Unpooled.copiedBuffer(ModbusConsts.HEARTBEAT, CharsetUtil.UTF_8));

    private final Map<String, Integer> heartbeatIdleCounter = Maps.newConcurrentMap();

    private boolean isSlave = false;

    public ModbusHeartbeatHandler(boolean isSlave) {
        this.isSlave = isSlave;
    }

    public void channelRead(ChannelHandlerContext ctx, Object msg) throws Exception {
        String key = "" + ctx.channel().remoteAddress();
        this.heartbeatIdleCounter.put(key, Integer.valueOf(0));
        logger.debug(ctx.channel(), String.format("channelRead heartbeatIdle:%s", this.heartbeatIdleCounter.get(key)));
    }

    public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
        if (evt instanceof IdleStateEvent) {
            IdleStateEvent event = (IdleStateEvent) evt;
            if (!this.isSlave) {
                handleIdleStateEvent4Master(ctx, event);
            } else {
                handleIdleStateEvent4Salve(ctx, event);
            }
        } else {
            super.userEventTriggered(ctx, evt);
        }
    }

    protected void handleIdleStateEvent4Master(ChannelHandlerContext ctx, IdleStateEvent event) {
        IdleState state = event.state();
        switch (state) {
            case READER_IDLE:
                handleIdle4Master(ctx, state);
                break;
            case WRITER_IDLE:
                handleIdle4Master(ctx, state);
                break;
            case ALL_IDLE:
                handleIdle4Master(ctx, state);
                break;
        }
    }

    protected void handleIdleStateEvent4Salve(ChannelHandlerContext ctx, IdleStateEvent event) {
        String key = "" + ctx.channel().remoteAddress();
        if (!this.heartbeatIdleCounter.containsKey(key)) {
            this.heartbeatIdleCounter.put(key, Integer.valueOf(0));
        } else {
            this.heartbeatIdleCounter.put(key, Integer.valueOf(this.heartbeatIdleCounter.get(key).intValue() + 1));
        }
        if (this.heartbeatIdleCounter.get(key).intValue() >= 3) {
            IdleState state = event.state();
            switch (state) {
                case READER_IDLE:
                    handleIdle4Slave(ctx, state);
                    break;
                case WRITER_IDLE:
                    handleIdle4Slave(ctx, state);
                    break;
                case ALL_IDLE:
                    handleIdle4Slave(ctx, state);
                    break;
            }
        }
    }

    protected void handleIdle4Master(ChannelHandlerContext ctx, IdleState state) {
        logger.debug(ctx.channel(), String.format("idle4Master:%s", state));
        ctx.writeAndFlush(HEARTBEAT.duplicate()).addListener(ChannelFutureListener.CLOSE_ON_FAILURE);
    }

    protected void handleIdle4Slave(ChannelHandlerContext ctx, IdleState state) {
        String key = "" + ctx.channel().remoteAddress();
        logger.debug(ctx.channel(), String.format("idle4Slave:%s,%s", state, this.heartbeatIdleCounter.get(key)));
        ctx.channel().close();
    }
}
