package com.github.zengfr.easymodbus4j.handler;

import com.github.zengfr.easymodbus4j.ModbusConfs;
import com.github.zengfr.easymodbus4j.cache.ModebusFrameCache;
import com.github.zengfr.easymodbus4j.cache.ModebusFrameCacheFactory;
import com.github.zengfr.easymodbus4j.event.SyncStateEvent;
import com.github.zengfr.easymodbus4j.func.AbstractRequest;
import com.github.zengfr.easymodbus4j.func.response.ErrorFunctionResponse;
import com.github.zengfr.easymodbus4j.logging.ChannelLogger;
import com.github.zengfr.easymodbus4j.protocol.ModbusFunction;
import com.github.zengfr.easymodbus4j.protocol.tcp.ModbusFrame;
import com.github.zengfr.easymodbus4j.util.ModbusTransactionIdUtil;
import io.netty.channel.Channel;
import io.netty.channel.ChannelHandlerContext;
import io.netty.util.Attribute;
import io.netty.util.AttributeKey;

public abstract class ModbusResponseHandler extends ModbusInboundHandler {
    private static final ChannelLogger logger = ChannelLogger.getLogger(ModbusResponseHandler.class);

    private boolean isCacheResponse = false;

    private int responseFrameIgnoreLengthThreshold = 0;

    public ModbusResponseHandler(boolean isCacheResponse) {
        this(isCacheResponse, ModbusConfs.RESPONS_EFRAME_IGNORE_LENGTH_THRESHOLD);
    }

    public ModbusResponseHandler(boolean isCacheResponse, int responseFrameIgnoreLengthThreshold) {
        this.isCacheResponse = isCacheResponse;
        this.responseFrameIgnoreLengthThreshold = responseFrameIgnoreLengthThreshold;
    }

    protected void channelRead0(ChannelHandlerContext ctx, ModbusFrame response) throws Exception {
        logger.debug(ctx.channel(), "channelRead0");
        cacheResponse(response, ctx.channel());
        boolean isFrameIgnored = isFrameIgnored(response);
        AttributeKey<Boolean> key = AttributeKey.valueOf("isFrameIgnored");
        Attribute<Boolean> attr = ctx.channel().attr(key);
        attr.set(Boolean.valueOf(isFrameIgnored));
        if (!isFrameIgnored) {
            boolean success = processResponseFrame(ctx.channel(), response);
            if (success) {
                ctx.fireUserEventTriggered(new SyncStateEvent(Short.valueOf(response.getFunction().getFunctionCode()), true, false));
            } else {
                logger.debug(ctx.channel(), "success:" + success);
            }
        } else {
            logger.debug(ctx.channel(), "isFrameIgnored:" + isFrameIgnored);
        }
    }

    protected boolean isFrameIgnored(ModbusFrame response) {
        int ignoreLengthThreshold = this.responseFrameIgnoreLengthThreshold;
        return (ignoreLengthThreshold > 0 && response.getHeader().getLength() > ignoreLengthThreshold);
    }

    protected void cacheResponse(ModbusFrame response, Channel channel) {
        if (this.isCacheResponse && response != null) {
            Short funcCode = Short.valueOf(response.getFunction().getFunctionCode());
            int respTransactionIdentifier = response.getHeader().getTransactionIdentifier();
            ModebusFrameCache responseCache = ModebusFrameCacheFactory.getInstance().getResponseCache(funcCode);
            responseCache.put(Integer.valueOf(respTransactionIdentifier), response);

            int respTransactionIdentifier2 = ModbusTransactionIdUtil.getTransactionIdByRemote(channel);
            responseCache.put(respTransactionIdentifier2, response);
        }
    }

    protected boolean processResponseFrame(Channel channel, ModbusFrame respFrame) {
        boolean success = false;
        int respTransactionIdentifier = respFrame.getHeader().getTransactionIdentifier();
        int reqTransactionIdentifier = getReqTransactionIdByRespTransactionId(respTransactionIdentifier);
        if (respTransactionIdentifier < 0 || reqTransactionIdentifier < 0)
            respTransactionIdentifier = reqTransactionIdentifier = ModbusTransactionIdUtil.getTransactionIdByRemote(channel);
        short funcCode = respFrame.getFunction().getFunctionCode();
        ModbusFrame reqFrame = getRequestCache(reqTransactionIdentifier, funcCode);
        AbstractRequest reqFunc = null;
        if (reqFrame != null) {
            reqFunc = (AbstractRequest) reqFrame.getFunction();
        } else {
            logger.warn(channel, String.format("req is null:%s;%s;%s;%s", Integer.valueOf(reqTransactionIdentifier), Integer.valueOf(respTransactionIdentifier), getRequestCache(funcCode).keySet(), respFrame));
        }
        short unitId = respFrame.getHeader().getUnitIdentifier();
        ModbusFunction respFunc = respFrame.getFunction();
        if (reqFrame == null || reqFrame.getHeader().getUnitIdentifier() == unitId) {
            success = processResponseFrame(channel, unitId, reqFunc, respFunc);
            if (success)
                removeRequestCache(reqTransactionIdentifier, funcCode);
        }
        return success;
    }

    protected ModebusFrameCache getRequestCache(short funcCode) {
        return ModebusFrameCacheFactory.getInstance().getRequestCache(Short.valueOf(funcCode));
    }

    protected ModbusFrame getRequestCache(int reqTransactionIdentifier, short funcCode) {
        ModbusFrame reqFrame = getRequestCache(funcCode).get(Integer.valueOf(reqTransactionIdentifier));
        return reqFrame;
    }

    protected void removeRequestCache(int reqTransactionIdentifier, short funcCode) {
        getRequestCache(funcCode).remove(Integer.valueOf(reqTransactionIdentifier));
    }

    protected ModebusFrameCache getResponseCache(short funcCode) {
        return ModebusFrameCacheFactory.getInstance().getResponseCache(Short.valueOf(funcCode));
    }

    public ModbusFrame getResponseCache(int respTransactionIdentifier, short funcCode) throws Exception {
        ModbusFrame frame;
        long timeoutTime = System.currentTimeMillis() + ModbusConfs.SYNC_RESPONSE_TIMEOUT;
        ModebusFrameCache responseCache = getResponseCache(funcCode);
        do {
            frame = responseCache.get(respTransactionIdentifier);
            if (frame == null){
                Thread.sleep(22L);
            }
        } while (frame == null && timeoutTime - System.currentTimeMillis() > 0L);

        if (frame != null)
            responseCache.remove(respTransactionIdentifier);
        if (frame == null)
            throw new Exception(String.format("resp is null!(%s,%s,%s,%s)", ModbusConfs.SYNC_RESPONSE_TIMEOUT,respTransactionIdentifier, responseCache.size(), responseCache.keySet()));
        if (frame.getFunction() instanceof ErrorFunctionResponse)
            throw new Exception(((ErrorFunctionResponse) frame.getFunction()).getExceptionMessage());
        return frame;
    }

    protected abstract int getReqTransactionIdByRespTransactionId(int paramInt);

    protected abstract boolean processResponseFrame(Channel paramChannel, int paramInt, AbstractRequest paramAbstractRequest, ModbusFunction paramModbusFunction);
}
