package com.github.zengfr.easymodbus4j.sender;

import com.github.zengfr.easymodbus4j.ModbusConsts;
import com.github.zengfr.easymodbus4j.cache.ModebusFrameCacheFactory;
import com.github.zengfr.easymodbus4j.handler.ModbusChannelSyncHandler;
import com.github.zengfr.easymodbus4j.handler.ModbusResponseHandler;
import com.github.zengfr.easymodbus4j.protocol.ModbusFunction;
import com.github.zengfr.easymodbus4j.protocol.tcp.ModbusFrame;
import com.github.zengfr.easymodbus4j.protocol.tcp.ModbusHeader;
import com.github.zengfr.easymodbus4j.util.ModbusTransactionIdUtil;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelFutureListener;

public abstract class AbstractChannelSender {
    private final short protocolIdentifier;

    private final short unitIdentifier;

    private final Channel channel;

    public Channel getChannel() {
        return this.channel;
    }

    public AbstractChannelSender(Channel channel) {
        this(channel, ModbusConsts.DEFAULT_UNIT_IDENTIFIER, (short) 0);
    }

    public AbstractChannelSender(Channel channel, short unitIdentifier) {
        this(channel, unitIdentifier, (short) 0);
    }

    public AbstractChannelSender(Channel channel, short unitId, short protocolIdentifier) {
        this.channel = channel;
        this.unitIdentifier = unitId;
        this.protocolIdentifier = protocolIdentifier;
    }

    protected int callModbusFunction(ModbusFunction function) throws NullPointerException {
        if (getChannel() == null)
            throw new NullPointerException("channel");
        ModbusFrame frame = buildModbusFrame(function);
        return sendModbusFrame(frame);
    }

    protected ModbusFrame buildModbusFrame(ModbusFunction function) {
        ModbusHeader header = buildModbusHeader(function);
        return new ModbusFrame(header, function);
    }

    protected ModbusHeader buildModbusHeader(ModbusFunction function) {
        int transactionId = ModbusTransactionIdUtil.calculateTransactionId();
        int pduLength = function.calculateLength();
        ModbusHeader header = new ModbusHeader(transactionId, this.protocolIdentifier, pduLength, this.unitIdentifier);
        return header;
    }

    public int sendModbusFrame(final ModbusFrame frame) {
        final Short funcCode = Short.valueOf(frame.getFunction().getFunctionCode());
        ModebusFrameCacheFactory.getInstance().getRequestCache(funcCode).put(frame);
        final int transactionId = ModbusTransactionIdUtil.getTransactionIdByRemote(this.channel);
        ChannelFutureListener channelFutureListener = new ChannelFutureListener() {
            public void operationComplete(ChannelFuture channelFuture) throws Exception {
                if (channelFuture.isSuccess()) {
                    ModebusFrameCacheFactory.getInstance().getRequestCache(funcCode).put(Integer.valueOf(transactionId), frame);
                } else {
                    channelFuture.channel().close();
                }
            }
        };
        Channel channel = getChannel();
        //channelSyncTryAcquire(channel, funcCode);
        channel.writeAndFlush(frame).addListener(channelFutureListener);
        return (frame.getHeader().getTransactionIdentifier() < 0) ? transactionId : frame.getHeader().getTransactionIdentifier();
    }

    protected boolean isChannelSync(Channel channel) {
        return (channel.pipeline().get("sync") != null);
    }

    protected void channelSyncTryAcquire(Channel channel, Short funcCode) {
        if (isChannelSync(channel)) {
            ModbusChannelSyncHandler handler = (ModbusChannelSyncHandler) channel.pipeline().get("sync");
            handler.tryAcquire(channel, funcCode);
        }
    }

    protected <V extends ModbusFunction> V callModbusFunctionSync(ModbusFunction function) throws Exception {
        callModbusFunction(function);
        short funcCode = function.getFunctionCode();
        int respTransactionIdentifier2 = ModbusTransactionIdUtil.getTransactionIdByRemote(channel);
        ModbusResponseHandler responseHandler = getResponseHandler(getChannel());
        return responseHandler == null ? null : (V) responseHandler.getResponseCache(respTransactionIdentifier2, funcCode).getFunction();
    }

    protected ModbusResponseHandler getResponseHandler(Channel channel) {
        return (ModbusResponseHandler) channel.pipeline().get("responseHandler");
    }
}
