package com.github.zengfr.easymodbus4j.sender.util;

import com.github.zengfr.easymodbus4j.common.util.IntArrayUtil;
import com.github.zengfr.easymodbus4j.common.util.ParseStringUtil;
import com.github.zengfr.easymodbus4j.protocol.ModbusFunction;
import com.github.zengfr.easymodbus4j.sender.ChannelSender;
import com.github.zengfr.easymodbus4j.sender.ChannelSenderFactory;
import io.netty.channel.Channel;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;

import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.List;

public class ModbusRequestSendUtil {
    public enum PriorityStrategy {
        Channel, Req
    }

    private static final InternalLogger logger = InternalLoggerFactory.getInstance(ModbusRequestSendUtil.class);

    public static void sendRequests(Collection<Channel> channels, List<String> reqs, boolean isAllUseAsync, int fixedDelay, PriorityStrategy strategy) {
        switch (strategy) {
            case Channel:
                sendRequestsByChannelFirst(channels, reqs, isAllUseAsync, fixedDelay);
            case Req:
                sendRequestsByReqFirst(channels, reqs, isAllUseAsync, fixedDelay);
                break;
        }
    }

    private static void sendRequestsByChannelFirst(Collection<Channel> channels, List<String> reqs, boolean isAllUseAsync, int fixedDelay) {
        for (Channel channel : channels) {
            if (channel == null || !channel.isActive() || !channel.isOpen() || !channel.isWritable())
                continue;
            ChannelSender sender = ChannelSenderFactory.getInstance().get(channel);
            for (String str : reqs) {
                try {
                    long startTime = System.currentTimeMillis();
                    String[] args = str.split("[;|]");
                    if (args.length >= 3) {
                        String funcString = args[0];
                        String addressString = args[1];
                        String value = args[2];
                        if (isAllUseAsync && !funcString.endsWith("Async"))
                            funcString = funcString + "Async";
                        if (!funcString.endsWith("Async")) {
                            sendSyncFunc(sender, funcString, addressString, value);
                        } else {
                            sendAsyncFunc(sender, funcString, addressString, value);
                        }
                        long endTime = System.currentTimeMillis();
                        long span = endTime - startTime;
                        if (fixedDelay - span > 0L)
                            Thread.sleep(fixedDelay - span);
                    }
                } catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException |
                         IllegalArgumentException | SecurityException | InterruptedException ex) {
                    logger.error("sendRequestsByChannelFirst", ex);
                }
            }
        }
    }

    private static void sendRequestsByReqFirst(Collection<Channel> channels, List<String> reqs, boolean isAllUseAsync, int fixedDelay) {
        for (String str : reqs) {
            long startTime = System.currentTimeMillis();
            for (Channel channel : channels) {
                try {
                    if (channel == null || !channel.isActive() || !channel.isOpen() || !channel.isWritable())
                        continue;
                    ChannelSender sender = ChannelSenderFactory.getInstance().get(channel);
                    String[] args = str.split("[;|]");
                    if (args.length >= 3) {
                        String funcString = args[0];
                        String addressString = args[1];
                        String value = args[2];
                        if (isAllUseAsync && !funcString.endsWith("Async"))
                            funcString = funcString + "Async";
                        if (!funcString.endsWith("Async")) {
                            sendSyncFunc(sender, funcString, addressString, value);
                            continue;
                        }
                        sendAsyncFunc(sender, funcString, addressString, value);
                    }
                } catch (NoSuchMethodException | InvocationTargetException | IllegalAccessException |
                         IllegalArgumentException | SecurityException ex) {
                    logger.error("sendRequestsByReqFirst", ex);
                }
            }
            long endTime = System.currentTimeMillis();
            long span = endTime - startTime;
            if (span < fixedDelay)
                try {
                    Thread.sleep(fixedDelay - span);
                } catch (InterruptedException ex) {
                    logger.error("sendRequestsByReqFirst", ex);
                }
        }
    }

    public static ModbusFunction sendSyncFunc(ChannelSender sender, String funcString, String addressStr, String value) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, NoSuchMethodException, SecurityException {
        ModbusFunction respFunction = null;
        if (!funcString.startsWith("#")) {
            Integer address = Integer.valueOf(addressStr);
            if ("TF".contains(value)) {
                respFunction = ModbusChannelSenderUtil.sendSyncFunc(sender, funcString, address.intValue(), value.equalsIgnoreCase("T"));
            } else if (value.contains(",")) {
                String[] values = value.split("[,]");
                if (value.contains("T") || value.contains("F")) {
                    respFunction = ModbusChannelSenderUtil.sendSyncFunc(sender, funcString, address.intValue(), ParseStringUtil.parseBooleanArray(values));
                } else {
                    respFunction = ModbusChannelSenderUtil.sendSyncFunc(sender, funcString, address.intValue(), IntArrayUtil.toIntArray(values));
                }
            } else {
                respFunction = ModbusChannelSenderUtil.sendSyncFunc(sender, funcString, address.intValue(), Integer.valueOf(value).intValue());
            }
        }
        return respFunction;
    }

    public static int sendAsyncFunc(ChannelSender sender, String funcString, String addressStr, String value) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, NoSuchMethodException, SecurityException {
        int transactionId = 0;
        if (!funcString.startsWith("#")) {
            Integer address = Integer.valueOf(addressStr);
            if (!funcString.endsWith("Async"))
                funcString = funcString + "Async";
            if ("TF".contains(value)) {
                transactionId = ModbusChannelSenderUtil.sendAsyncFunc(sender, funcString, address.intValue(), value.equalsIgnoreCase("T"));
            } else if (value.contains(",")) {
                String[] values = value.split("[,]");
                if (value.contains("T") || value.contains("F")) {
                    transactionId = ModbusChannelSenderUtil.sendAsyncFunc(sender, funcString, address.intValue(), ParseStringUtil.parseBooleanArray(values));
                } else {
                    transactionId = ModbusChannelSenderUtil.sendAsyncFunc(sender, funcString, address.intValue(), IntArrayUtil.toIntArray(values));
                }
            } else {
                transactionId = ModbusChannelSenderUtil.sendAsyncFunc(sender, funcString, address.intValue(), Integer.valueOf(value).intValue());
            }
        }
        return transactionId;
    }
}
