package com.github.zengfr.easymodbus4j.server;

import com.github.zengfr.easymodbus4j.channel.ChannelManager;
import com.github.zengfr.easymodbus4j.channel.ChannelManagerImpl;
import com.github.zengfr.easymodbus4j.handler.ModbusInboundHandler;
import com.github.zengfr.easymodbus4j.handler.ModbusRequestHandler;
import com.github.zengfr.easymodbus4j.handler.ModbusResponseHandler;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.util.concurrent.DefaultThreadFactory;
import io.netty.util.concurrent.GenericFutureListener;
import io.netty.util.internal.logging.InternalLogger;
import io.netty.util.internal.logging.InternalLoggerFactory;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Set;
import java.util.concurrent.ThreadFactory;

public abstract class ModbusServer extends ChannelManagerImpl {
    public enum CONNECTION_STATES {
        listening, down, clientsConnected
    }

    protected static final InternalLogger logger = InternalLoggerFactory.getInstance(ModbusServer.class);

    public static final String PROP_CONNECTIONSTATE = "connectionState";

    private final int port;

    private ServerBootstrap bootstrap;

    private CONNECTION_STATES connectionState = CONNECTION_STATES.down;

    private final transient PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    private Channel parentChannel;

    public ModbusServer(int port) {
        this.port = port;
    }

    protected void setup(ModbusRequestHandler reqhandler) throws Exception {
        reqhandler.setChannelManager(this);
        ChannelInitializer<SocketChannel> handler = getChannelInitializer(reqhandler);
        setup(handler);
    }

    protected void setup(ModbusResponseHandler resphandler) throws Exception {
        resphandler.setChannelManager(this);
        ChannelInitializer<SocketChannel> handler = getChannelInitializer(resphandler);
        setup(handler);
    }

    public void setup(ChannelInitializer<SocketChannel> handler) throws Exception {
        init(handler);
    }

    protected abstract ChannelInitializer<SocketChannel> getChannelInitializer(ModbusInboundHandler paramModbusInboundHandler);

    protected void init(ChannelInitializer<SocketChannel> handler) throws Exception {
        try {
            int threads = Math.max(3, Runtime.getRuntime().availableProcessors() * 2 - 1);
            final NioEventLoopGroup bossGroup = new NioEventLoopGroup(threads, new DefaultThreadFactory("boss", false));
            final NioEventLoopGroup workerGroup = new NioEventLoopGroup(threads, new DefaultThreadFactory("worker", false));
            this.bootstrap = new ServerBootstrap();
            this.bootstrap.group(bossGroup,workerGroup).channel(NioServerSocketChannel.class).option(ChannelOption.SO_REUSEADDR, Boolean.valueOf(true)).option(ChannelOption.SO_BACKLOG, Integer.valueOf(1024))
                    .childOption(ChannelOption.TCP_NODELAY, Boolean.valueOf(true)).childOption(ChannelOption.SO_KEEPALIVE, Boolean.valueOf(true)).childHandler(handler);
            logger.info("bind port:" + this.port);
            this.parentChannel = this.bootstrap.bind(this.port).sync().channel();
            setConnectionState(CONNECTION_STATES.listening);
            this.parentChannel.closeFuture().addListener(new GenericFutureListener<ChannelFuture>() {
                public void operationComplete(ChannelFuture future) throws Exception {
                    workerGroup.shutdownGracefully();
                    bossGroup.shutdownGracefully();
                    ModbusServer.this.setConnectionState(CONNECTION_STATES.down);
                }
            });
        } catch (Exception ex) {
            setConnectionState(CONNECTION_STATES.down);
            logger.error("init", ex);
            throw new Exception("ConnectionException:" + ex.getLocalizedMessage());
        }
    }

    public CONNECTION_STATES getConnectionState() {
        return this.connectionState;
    }

    public void setConnectionState(CONNECTION_STATES connectionState) {
        CONNECTION_STATES oldConnectionState = this.connectionState;
        this.connectionState = connectionState;
        this.propertyChangeSupport.firePropertyChange("connectionState", oldConnectionState, connectionState);
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        this.propertyChangeSupport.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        this.propertyChangeSupport.removePropertyChangeListener(listener);
    }

    public void close() {
        if (this.parentChannel != null)
            this.parentChannel.close().awaitUninterruptibly();
        super.close();
    }

    public void removeChannel(Channel channel) {
        super.removeChannel(channel);
        setConnectionState(CONNECTION_STATES.clientsConnected);
    }

    public void addChannel(Channel channel) {
        super.addChannel(channel);
        setConnectionState(CONNECTION_STATES.clientsConnected);
    }

    public Channel getChannelByIp(String channelAddress) {
        Set<String> keys = this.channels.keySet();
        for (String key : keys) {
            if (key != null && key.indexOf(channelAddress) < 0)
                return this.channels.get(key);
        }
        return null;
    }
}
