package tofly.modbus_gateway2;

import com.github.zengfr.easymodbus4j.ModbusConfs;
import com.github.zengfr.easymodbus4j.handle.impl.ModbusMasterResponseHandler;
import com.github.zengfr.easymodbus4j.handler.ModbusResponseHandler;
import com.github.zengfr.easymodbus4j.processor.ModbusMasterResponseProcessor;
import com.github.zengfr.easymodbus4j.sender.ChannelSender;
import com.github.zengfr.easymodbus4j.server.ModbusServer4TcpMaster;
import io.netty.channel.Channel;
import io.netty.util.Attribute;
import io.netty.util.AttributeKey;
import org.springframework.boot.ApplicationRunner;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.boot.context.ApplicationPidFileWriter;
import org.springframework.cloud.openfeign.EnableFeignClients;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.annotation.Bean;
import org.springframework.scheduling.annotation.EnableAsync;
import org.springframework.scheduling.annotation.EnableScheduling;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.scheduling.support.PeriodicTrigger;
import tofly.modbus_gateway2.config.ModbusServerProperties;
import tofly.modbus_gateway2.log.Log;
import tofly.modbus_gateway2.log.LogFactory;
import tofly.modbus_gateway2.service.MonitorService;
import tofly.modbus_gateway2.service.StoreService;
import tofly.modbus_gateway2.util.DataTypeUtil;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static tofly.modbus_gateway2.DataType.valueOf;
import static tofly.modbus_gateway2.FirstFrameDecoder.ATTR_KEY_DEVICE_ID;

@SpringBootApplication
@EnableScheduling
@EnableFeignClients
@EnableAsync
public class Application {

    private static final Log log = LogFactory.getLog(Application.class);

    //private static final String SPRING_CONFIG_NAME_KEY = "--spring.config.name";
    //private static final String DEFAULT_SPRING_CONFIG_PARAM = SPRING_CONFIG_NAME_KEY + "=" + "modbus_gateway2";
    private static MonitorService monitorService = null;

    public static void main(String[] args) {
        SpringApplication springApplication = new SpringApplication(Application.class);
        springApplication.addListeners(new ApplicationPidFileWriter());
        springApplication.run(args);
        //springApplication.run(updateArguments(args));

        Runtime.getRuntime().addShutdownHook(new Thread(() -> {
            // 在JVM关闭之前执行收尾工作
            // 注意事项：
            // 1.在这里执行的动作不能耗时太久
            // 2.不能在这里再执行注册，移除关闭钩子的操作
            // 3 不能在这里调用System.exit()
            log.info("-----do shutdown hook-----");
            if (monitorService != null) {
                monitorService.resetAllState();
            }
        }));
    }

    @Bean
    public ModbusServer4TcpMaster getModbusServer4TcpMaster(ModbusServerProperties modbusServerProperties,
                                                            StoreService storeService,
                                                            MonitorService monitorService,
                                                            ApplicationEventPublisher eventPublisher) throws Exception {
        int port = modbusServerProperties.getPort();
        short transactionIdentifierOffset = 0;
        ModbusConfs.MASTER_SHOW_DEBUG_LOG = false;
        ModbusConfs.FRAME_CACHE_EXPIRE = 15000;
        ModbusConfs.SYNC_TRYACQUIRE_TIMEOUT_Millis = 15000;
        ModbusConfs.SYNC_RESPONSE_TIMEOUT = 15000;
        ModbusMasterResponseProcessor masterProcessor = new GwModbusMasterResponseProcessor(
                transactionIdentifierOffset,
                modbusServerProperties.isStore() ? storeService : null,
                modbusServerProperties.isStore() ? monitorService : null);
        ModbusResponseHandler responseHandler = new ModbusMasterResponseHandler(masterProcessor);
        //return ModbusServerTcpFactory.getInstance().createServer4Master(port, responseHandler);

        GwModbusServer4TcpMaster modbusServer4Master = new GwModbusServer4TcpMaster(port, eventPublisher);
        new GwModbusTcpChannelInitializer(false, responseHandler, modbusServer4Master, eventPublisher, modbusServerProperties);
        responseHandler.setChannelManager(modbusServer4Master);

        return modbusServer4Master;
    }

    public final static AttributeKey<ModbusServerProperties.DeviceConfiguration> ATTR_KEY_DEVICE = AttributeKey.valueOf("device");

    @Bean
    public ApplicationRunner applicationRunner(ModbusServerProperties modbusServerProperties,
                                               ModbusServer4TcpMaster modbusServer4TcpMaster,
                                               MonitorService monitorService) {
        boolean store = modbusServerProperties.isStore();
        if (store) {
            Application.monitorService = monitorService;
            try {
                monitorService.resetAllState();
            } catch (Exception e) {
                log.warn("初始化异常", e);
            }
        }
        return args -> {
            modbusServerProperties.getSlaveMappings().forEach((k, v) -> {
                int period = v.getPeriod();
                PeriodicTrigger trigger = new PeriodicTrigger(period);
                trigger.setInitialDelay(3000);
                threadPoolTaskScheduler().schedule(() -> {
                    try {
                        Collection<Channel> channels = modbusServer4TcpMaster.getChannels();
                        for (Channel channel : channels) {
                            if (channel == null || (!channel.isActive()) || !channel.isOpen() || !channel.isWritable())
                                continue;
                            Attribute<String> attrDeviceId = channel.attr(ATTR_KEY_DEVICE_ID);
                            String deviceId = attrDeviceId.get();
                            if (deviceId == null) continue;
                            log.debug("读取数据:" + deviceId);
                            try {
                                ModbusServerProperties.DeviceConfiguration deviceConfiguration = channel.attr(ATTR_KEY_DEVICE).get();
                                Map<String, ModbusServerProperties.FieldConfiguration> fieldMappings = deviceConfiguration.getFieldMappings();

                                Set<String> fields = fieldMappings.keySet();
                                if (fields.isEmpty()) continue;

                                List<Integer> list = fields.stream()
                                        .map(e -> getQuantityOfRegisters(e)).collect(Collectors.toList());
                                Integer quantityOfRegisters = list.stream()
                                        .max(Integer::compareTo).get();

                                if (quantityOfRegisters == 0) continue;
                                short unitIdentifier = deviceConfiguration.getUnitIdentifier();
                                ChannelSender sender = new ChannelSender(channel, unitIdentifier);

                                int startAddress = 0;

                                Integer functionCode = deviceConfiguration.getFunctionCode();
                                if (functionCode == null || 3 == functionCode) {
                                    sender.readInputRegistersAsync(startAddress, quantityOfRegisters);
                                } else if (4 == functionCode) {
                                    sender.readHoldingRegistersAsync(startAddress, quantityOfRegisters);
                                }
                            } catch (Exception e) {
                                log.error(String.format("读取数据异常:%s", deviceId), e);
                            }
                        }
                    } catch (Throwable e) {
                        log.error("未知异常", e);
                    }
                }, trigger);
            });

        };
    }

    private int getQuantityOfRegisters(String e) {
        String[] parts = e.split("_");
        if (parts.length < 2) throw new IllegalArgumentException();
        Integer startAddress = Integer.parseInt(parts[0]);
        String dataType = parts[1];
        DataType type = valueOf(dataType.replaceAll("-", "_").toUpperCase());
        int byteLen = DataTypeUtil.getByteLen(type);
        int y = 2;
        return startAddress + (byteLen / y + (byteLen % y != 0 ? 1 : 0));
    }

    @Bean(name = "threadPoolTaskScheduler")
    public ThreadPoolTaskScheduler threadPoolTaskScheduler() {
        ThreadPoolTaskScheduler threadPoolTaskScheduler = new ThreadPoolTaskScheduler();
        //threadPoolTaskScheduler.setDaemon(true);
        int availableProcessors = Runtime.getRuntime().availableProcessors();
        threadPoolTaskScheduler.setPoolSize(availableProcessors * 2);
        threadPoolTaskScheduler.setThreadNamePrefix("Worker:");
        threadPoolTaskScheduler.setWaitForTasksToCompleteOnShutdown(true);
        threadPoolTaskScheduler.setAwaitTerminationSeconds(30);
        threadPoolTaskScheduler.initialize();
        return threadPoolTaskScheduler;
    }


//    private static String[] updateArguments(String[] args) {
//        if (Arrays.stream(args).noneMatch(arg -> arg.startsWith(SPRING_CONFIG_NAME_KEY))) {
//            String[] modifiedArgs = new String[args.length + 1];
//            System.arraycopy(args, 0, modifiedArgs, 0, args.length);
//            modifiedArgs[args.length] = DEFAULT_SPRING_CONFIG_PARAM;
//            return modifiedArgs;
//        }
//        return args;
//    }
}