package tofly.modbus_gateway2;

import com.github.zengfr.easymodbus4j.server.ModbusServer4TcpMaster;
import io.netty.buffer.ByteBuf;
import io.netty.channel.Channel;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.ByteToMessageDecoder;
import io.netty.util.Attribute;
import io.netty.util.AttributeKey;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.ApplicationEventPublisher;
import tofly.modbus_gateway2.config.ModbusServerProperties;
import tofly.modbus_gateway2.event.OnlineEvent;

import java.nio.charset.Charset;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static tofly.modbus_gateway2.Application.ATTR_KEY_DEVICE;
import static tofly.modbus_gateway2.HeartbeatFrameDecoder.ATTR_KEY_ALIVE_LAST_TIME;

@Slf4j
public class FirstFrameDecoder extends ByteToMessageDecoder {

    public final static AttributeKey<String> ATTR_KEY_DEVICE_ID = AttributeKey.valueOf("device_id");

    private final static Pattern pattern = Pattern.compile("XMYN(.*)\\s*");
    private final ModbusServer4TcpMaster modbusServer4TcpMaster;
    private final ApplicationEventPublisher eventPublisher;
    private final Map<String, ModbusServerProperties.DeviceConfiguration> deviceMappings;

    public FirstFrameDecoder(ModbusServer4TcpMaster modbusServer4TcpMaster, ApplicationEventPublisher eventPublisher, ModbusServerProperties modbusServerProperties) {
        this.modbusServer4TcpMaster = modbusServer4TcpMaster;
        this.eventPublisher = eventPublisher;
        Collection<ModbusServerProperties.SlaveConfiguration> values = modbusServerProperties.getSlaveMappings().values();
        this.deviceMappings = new LinkedHashMap<>();
        values.forEach(e -> deviceMappings.putAll(e.getDeviceMappings()));
    }

    @Override
    protected void decode(ChannelHandlerContext ctx, ByteBuf byteBuf, List<Object> list) throws Exception {

        Attribute<String> attrDeviceId = ctx.channel().attr(ATTR_KEY_DEVICE_ID);
        int readableBytes = byteBuf.readableBytes();
        if (attrDeviceId.get() == null) {
            CharSequence charSequence = byteBuf.readCharSequence(readableBytes, Charset.forName("US-ASCII"));
            String value = charSequence.toString();
            //value = "XMYN08370003011308370030";
            Matcher matcher = pattern.matcher(value);
            if (matcher.find()) {
                String deviceId = matcher.group(1);

                Channel onlineChannel = getOnlineChannel(deviceId);
                if (onlineChannel != null) {
                    onlineChannel.attr(ATTR_KEY_DEVICE_ID).set(null);
                    ChannelFuture close = onlineChannel.close();
                    close.addListener(future -> {
                        log.info("重新连接:{}", deviceId);
                        attrDeviceId.set(deviceId);
                        ctx.channel().attr(ATTR_KEY_ALIVE_LAST_TIME).set(System.currentTimeMillis());
                        ctx.channel().attr(ATTR_KEY_DEVICE).set(deviceMappings.get(deviceId));
                    });
                } else {
                    if (deviceMappings.containsKey(deviceId)) {
                        attrDeviceId.set(deviceId);
                        ctx.channel().attr(ATTR_KEY_ALIVE_LAST_TIME).set(System.currentTimeMillis());
                        ctx.channel().attr(ATTR_KEY_DEVICE).set(deviceMappings.get(deviceId));
                        eventPublisher.publishEvent(new OnlineEvent(deviceId));
                    } else {
                        log.warn("设备未配置:{}", deviceId);
                    }
                }
            } else {
                log.warn("设备不匹配:{}", value);
            }
        } else {
            ByteBuf frame = byteBuf.readBytes(readableBytes);
            list.add(frame);
        }
    }

    private Channel getOnlineChannel(String deviceId) throws InterruptedException {
        Collection<Channel> channels = modbusServer4TcpMaster.getChannels();
        Map<String, Channel> channelMap = channels.stream()
                .filter(e -> e.attr(ATTR_KEY_DEVICE_ID).get() != null)
                .collect(Collectors.toMap(e -> e.attr(ATTR_KEY_DEVICE_ID).get(), v -> v));
        return channelMap.get(deviceId);

    }
}
