package tofly.modbus_gateway2;

import com.github.zengfr.easymodbus4j.func.AbstractRequest;
import com.github.zengfr.easymodbus4j.func.response.ReadHoldingRegistersResponse;
import com.github.zengfr.easymodbus4j.func.response.ReadInputRegistersResponse;
import com.github.zengfr.easymodbus4j.processor.AbstractModbusProcessor;
import com.github.zengfr.easymodbus4j.processor.ModbusMasterResponseProcessor;
import com.github.zengfr.easymodbus4j.protocol.ModbusFunction;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.Channel;
import io.netty.util.Attribute;
import io.netty.util.AttributeKey;
import tofly.modbus_gateway2.config.ModbusServerProperties;
import tofly.modbus_gateway2.log.Log;
import tofly.modbus_gateway2.log.LogFactory;
import tofly.modbus_gateway2.service.MonitorService;
import tofly.modbus_gateway2.service.StoreService;
import tofly.modbus_gateway2.util.DataTypeUtil;
import tofly.modbus_gateway2.util.JsonUtil;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.net.InetSocketAddress;
import java.util.LinkedHashMap;
import java.util.Map;

import static tofly.modbus_gateway2.Application.ATTR_KEY_DEVICE;
import static tofly.modbus_gateway2.DataType.valueOf;
import static tofly.modbus_gateway2.FirstFrameDecoder.ATTR_KEY_DEVICE_ID;

public class GwModbusMasterResponseProcessor extends AbstractModbusProcessor implements ModbusMasterResponseProcessor {

    public final static AttributeKey<Long> ATTR_KEY_REFRESH_LAST_TIME = AttributeKey.valueOf("refresh_last_time");
    /**
     * 日志
     */
    private static final Log log = LogFactory.getLog(GwModbusMasterResponseProcessor.class);
    private final StoreService storeService;

    private final MonitorService monitorService;

    public GwModbusMasterResponseProcessor(short transactionIdentifierOffset, StoreService storeService, MonitorService monitorService) {
        super(transactionIdentifierOffset, false);
        this.storeService = storeService;
        this.monitorService = monitorService;
    }

    public boolean processResponseFrame(Channel channel, int unitId, AbstractRequest reqFunc, ModbusFunction respFunc) {
        InetSocketAddress inetSocketAddress = (InetSocketAddress) channel.remoteAddress();
        String deviceAddress = inetSocketAddress.getAddress().getHostAddress();
        Attribute<String> attrDeviceId = channel.attr(ATTR_KEY_DEVICE_ID);
        String deviceId = attrDeviceId.get();
        if (deviceId == null) {
            log.warn("设备ID为空");
        } else {
            if (respFunc instanceof ReadHoldingRegistersResponse || respFunc instanceof ReadInputRegistersResponse) {
                ModbusServerProperties.DeviceConfiguration deviceConfiguration = channel.attr(ATTR_KEY_DEVICE).get();
                if (deviceConfiguration != null) {
                    Map<String, Object> dataMap = new LinkedHashMap<>();
                    Map<String, ModbusServerProperties.FieldConfiguration> fieldMappings = deviceConfiguration.getFieldMappings();
                    fieldMappings.forEach((k, v) -> {
                        Object value;
                        if ((value = getValue(k, v, respFunc)) != null) dataMap.put(v.getName(), value);
                    });
                    if (dataMap.isEmpty()) return false;
                    if (log.isDebugEnabled()) {
                        log.debug(deviceId + ">>DATA:" + JsonUtil.toString(dataMap));
                    }
                    channel.attr(ATTR_KEY_REFRESH_LAST_TIME).set(System.currentTimeMillis());
                    if (storeService != null) {
                        storeService.store(deviceAddress, deviceId, dataMap);
                    }
                    if (monitorService != null) {
                        monitorService.storeData(deviceId, JsonUtil.toString(dataMap));
                    }
                    return true;
                } else {
                    log.warn("未读取到设备配置:" + deviceId);
                    return false;
                }
            } else {
                log.warn("未知响应:" + deviceId);
            }
        }
        return false;
        //log.debug(String.format("success:%s", success));

    }

    private Object getValue(String fieldKey, ModbusServerProperties.FieldConfiguration fieldConfig, ModbusFunction resp) {
        String[] parts = fieldKey.split("_");
        if (parts.length < 2) throw new IllegalArgumentException();
        Integer startAddress = Integer.parseInt(parts[0]);
        String dataType = parts[1];
        ByteBuf byteBuf = resp.encode();
        Object result = null;
        DataType dt = valueOf(dataType.replaceAll("-", "_").toUpperCase());
        int byteLen = DataTypeUtil.getByteLen(dt);
        int readableBytes = byteBuf.readableBytes();
        int index = 2 + 2 * startAddress;
        if (index >= readableBytes || index + byteLen > readableBytes) {
            return null;
        }
        ByteBuf dataByteBuf = byteBuf.slice(index, byteLen);
        ByteBuf newByteBuf = null;
        switch (dt) {
            case SIGNED:
            case SIGNED_AB:
                result = dataByteBuf.readShort();
                break;
            case SIGNED_BA:
                result = dataByteBuf.readShortLE();
                break;
            case UNSIGNED_AB:
                result = dataByteBuf.readUnsignedShort();
                break;
            case UNSIGNED_BA:
                result = dataByteBuf.readUnsignedShortLE();
                break;
            case LONG_ABCD:
                result = dataByteBuf.readInt();
                break;
            case LONG_CDAB:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(2, newByteBuf, 2);
                dataByteBuf.getBytes(0, newByteBuf, 2);
                result = newByteBuf.readInt();
                newByteBuf.release();
                break;
            case LONG_BADC:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(2, newByteBuf, 2);
                dataByteBuf.getBytes(0, newByteBuf, 2);
                result = newByteBuf.readIntLE();
                newByteBuf.release();
                break;
            case LONG_DCBA:
                result = dataByteBuf.readIntLE();
                break;
            case FLOAT_ABCD:
                result = dataByteBuf.readFloat();
                break;
            case FLOAT_CDAB:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(2, newByteBuf, 2);
                dataByteBuf.getBytes(0, newByteBuf, 2);
                result = newByteBuf.readFloat();
                newByteBuf.release();
                break;
            case FLOAT_BADC:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(2, newByteBuf, 2);
                dataByteBuf.getBytes(0, newByteBuf, 2);
                result = newByteBuf.readFloatLE();
                newByteBuf.release();
                break;
            case FLOAT_DCBA:
                result = dataByteBuf.readFloatLE();
                break;
            case DOUBLE_ABCDEFGH:
                result = newByteBuf.readDouble();
                break;
            case DOUBLE_GHEFCDAB:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(1, newByteBuf, 1);
                dataByteBuf.getBytes(0, newByteBuf, 1);
                dataByteBuf.getBytes(3, newByteBuf, 1);
                dataByteBuf.getBytes(2, newByteBuf, 1);
                dataByteBuf.getBytes(5, newByteBuf, 1);
                dataByteBuf.getBytes(4, newByteBuf, 1);
                dataByteBuf.getBytes(7, newByteBuf, 1);
                dataByteBuf.getBytes(6, newByteBuf, 1);
                result = newByteBuf.readDoubleLE();
                newByteBuf.release();
                break;
            case DOUBLE_BADCFEHG:
                newByteBuf = Unpooled.buffer(readableBytes);
                dataByteBuf.getBytes(7, newByteBuf, 1);
                dataByteBuf.getBytes(6, newByteBuf, 1);
                dataByteBuf.getBytes(5, newByteBuf, 1);
                dataByteBuf.getBytes(4, newByteBuf, 1);
                dataByteBuf.getBytes(3, newByteBuf, 1);
                dataByteBuf.getBytes(2, newByteBuf, 1);
                dataByteBuf.getBytes(1, newByteBuf, 1);
                dataByteBuf.getBytes(0, newByteBuf, 1);
                result = newByteBuf.readDoubleLE();
                newByteBuf.release();
                break;
            case DOUBLE_HGFEDCBA:
                result = dataByteBuf.readDoubleLE();
                break;
        }
        dataByteBuf.release();
        int fieldType = fieldConfig.getType();
        int digits = fieldConfig.getDigits();
        int intValue = 0;
        if (fieldType != 0) {
            intValue = new BigDecimal(String.valueOf(result)).intValue();
        }
        if (fieldType > 0) {
            int value = intValue;
            return new BigDecimal(calc(value, fieldType))
                    .setScale(digits, RoundingMode.HALF_UP).doubleValue();
        } else if (fieldType == -1) {
            int value = intValue;
            int lower = fieldConfig.getLower();
            int upper = fieldConfig.getUpper();
            int gcLower = fieldConfig.getGcLower();
            int gcUpper = fieldConfig.getGcUpper();
            return new BigDecimal(calc(value, upper, lower, gcUpper, gcLower))
                    .setScale(digits, RoundingMode.HALF_UP).doubleValue();
        }

        return result;
    }

    private static double calc(int value, int fieldType) {

        int upper = 2;
        int lower = 0;
        int gcUpper = 5000;
        int gcLower = 50;
        if (fieldType == 1) {
            upper = 1;
            lower = 0;
        } else if (fieldType == 2) {
            upper = 2;
            lower = 0;
        } else if (fieldType == 3) {
            upper = 5;
            lower = 0;
        } else if (fieldType == 4) {
            upper = 14;
            lower = 0;
        }

        return calc(value, upper, lower, gcUpper, gcLower);
    }

    private static double calc(int value, int upper, int lower, int gcUpper, int gcLower) {
        //(量程上限-量程下限) * (工程量实时值-工程量下限) / (工程量上限-工程量下限)
        return (upper - lower) * (value - gcLower) * 1D / (gcUpper - gcLower);
    }

//    public static void main(String[] args) {
//        System.out.println(calc(225, 2));
//        System.out.println(calc(676, 1));
//        System.out.println(calc(3000, 3));
//    }


//    private void pushData(String p_sDeviceAddress, String p_sToflyDeviceId, String p_sDeviceId, Map<String, Object> p_mData) {
//        List<String> jsonDatas = new ArrayList<>();
//        jsonDatas.add(JsonUtil.toString(p_mData));
//        log.info("解析后的报文：" + jsonDatas);
//        String sMqttUsername = g_dc_mqtt_username;
//        String sPassword = g_dc_mqtt_password;
//        try {
//            TransmitToMqtt.pushData(p_sToflyDeviceId,
//                    sMqttUsername,
//                    sPassword,
//                    jsonDatas);
//        } catch (MqttException ex) {
//            if (ex instanceof MqttSecurityException) {
//                log.info(p_sToflyDeviceId + "：" + ex.getMessage());
//                TransmitToMqtt.saveDeviceDiscovery(p_sDeviceAddress,
//                        "XRTU",
//                        p_sDeviceId,
//                        p_sToflyDeviceId, sMqttUsername,
//                        sPassword, jsonDatas.get(0));
//            } else {
//                log.warn("推送异常", ex);
//                //throw new RuntimeException(ex);
//            }
//        }
//    }


}
