package tofly.modbus_gateway2.service;

import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;
import tofly.modbus_gateway2.config.ModbusServerProperties;

import java.util.Date;
import java.util.List;
import java.util.Map;

@Service
@Slf4j
public class MonitorServiceImpl implements MonitorService {

    @Value("${spring.datasource.platform:oracle}")
    private String platform;

    @Autowired
    private JdbcTemplate jdbcTemplate;

    @Autowired
    private ModbusServerProperties modbusServerProperties;

    @Override
    @Async
    public void storeState(String deviceId, int state) {
        String table = modbusServerProperties.getMonitorTable();
        String sql = String.format("SELECT * FROM %s WHERE id = ?", table);
        List<Map<String, Object>> maps = jdbcTemplate.queryForList(sql, deviceId);
        Date now = new Date();
        if (maps.size() > 0) {
            update(deviceId, now, state);
        } else {
            save(deviceId, now, state);
        }
        saveHistory(deviceId, now, state);
    }

    @Override
    @Async
    public void storeData(String deviceId, String jsonData) {
        try {
            String table = modbusServerProperties.getMonitorTable();
            String sql = String.format("SELECT * FROM %s WHERE id = ?", table);
            List<Map<String, Object>> maps = jdbcTemplate.queryForList(sql, deviceId);
            if (maps.size() > 0) {
                update(deviceId, jsonData);
            } else {
                save(deviceId, jsonData);
            }
        } catch (Exception e) {
            log.warn(String.format("保存数据失败:%s,%s", deviceId, jsonData), e);
        }
    }

    @Override
    public void resetAllState() {
        Date now = new Date();
        List<String> onlineIds = getOnlineIds();
        int deviceState = 0;
        onlineIds.forEach(id -> {
            update(id, now, deviceState);
            saveHistory(id, now, deviceState);
        });

    }

    private List<String> getOnlineIds() {
        String table = modbusServerProperties.getMonitorTable();
        String sql = String.format("SELECT id FROM %s WHERE device_state = 1", table);
        List<String> ids = jdbcTemplate.query(sql, (rs, rowNum) -> rs.getString(1));
        return ids;
    }

    private void save(String deviceId, Date time, int state) {
        String table = modbusServerProperties.getMonitorTable();
        String sql = null;
        if (state == 0) {
            sql = String.format("INSERT INTO %s (id, last_offline_time, device_state) VALUES (?, ?, ?)", table);
        } else if (state == 1) {
            sql = String.format("INSERT INTO %s (id, last_online_time, device_state) VALUES (?, ?, ?)", table);
        } else {
            throw new IllegalArgumentException("无效参数");
        }
        jdbcTemplate.update(sql, deviceId, time, state);
    }

    private void saveHistory(String deviceId, Date time, int state) {
        String table = modbusServerProperties.getMonitorHistoryTable();
        String sql = null;
        if (state == 0) {
            sql = String.format("INSERT INTO %s (id, create_time, device_state) VALUES (?, ?, ?)", table);
        } else if (state == 1) {
            sql = String.format("INSERT INTO %s (id, create_time, device_state) VALUES (?, ?, ?)", table);
        } else {
            throw new IllegalArgumentException("无效参数");
        }
        jdbcTemplate.update(sql, deviceId, time, state);
    }

    private void update(String deviceId, Date time, int state) {

        String table = modbusServerProperties.getMonitorTable();
        String sql = null;
        if (state == 0) {
            sql = String.format("UPDATE %s SET device_state = ?, last_offline_time = ? WHERE id = ?", table);
        } else if (state == 1) {
            sql = String.format("UPDATE %s SET device_state = ?, last_online_time = ? WHERE id = ?", table);
        } else {
            throw new IllegalArgumentException("无效参数");
        }
        jdbcTemplate.update(sql, state, time, deviceId);

    }

    private void save(String deviceId, String jsonData) {
        String table = modbusServerProperties.getMonitorTable();
        String sql = String.format("INSERT INTO %s (id, last_refresh_time, last_data) VALUES (?, ?, ?)", table);
        ;
        Date now = new Date();
        jdbcTemplate.update(sql, deviceId, now, jsonData);
    }

    private void update(String deviceId, String jsonData) {
        String table = modbusServerProperties.getMonitorTable();
        String sql = String.format("UPDATE %s SET last_data = ?, last_refresh_time = ? WHERE id = ?", table);
        Date now = new Date();
        jdbcTemplate.update(sql, jsonData, now, deviceId);
    }


}
